-- Conversation Manager

--=== Local junk ===========================================================--
local _G = _G
local _

local function cMessage(msg)
	if (DEFAULT_CHAT_FRAME) then
		DEFAULT_CHAT_FRAME:AddMessage(msg);
	end
end

local function cColourString(r, g, b, bold)

	if bold then
		r = (r + 1) / 2;
		g = (g + 1) / 2;
		b = (b + 1) / 2;
	end

	return string.format("|cff%02x%02x%02x", r * 255, g * 255, b * 255);
end

local function cChatColourString(chattype, bold)
	if not ChatTypeInfo[chattype] then
		return "";
	end

	return cColourString(ChatTypeInfo[chattype].r, ChatTypeInfo[chattype].g, ChatTypeInfo[chattype].b, bold);
end

--===--

local cNumItems = 0;		-- Number of 'items' on the side list
local cSelected = 1;		-- Currently selected
local cItems = {};		-- Table of sidebar items
local cChannels = {};		-- Table of numbered channels we are currently in. Stored in lowercase
local cServerChannels = {};	-- Table of server channels
local lastScroll = 0 ;   	-- Last time we scrolled the history
local autoScroll = 1;		-- We auto scrolled back

local cChannelKludge = "";	-- Since CHAT_MSG_CHANNEL_NOTICE seems to fire BEFORE it actually removes you from the channel

local cFinishedLoading = nil;

local cDateStamp = "";		-- Constant datestamp generated per login

local RealmName = "";
local PlayerName = "";
local GuildName = "";



--=== Initialization =======================================================--

function Converse_OnLoad()
	-- So ESC can close it...
	tinsert(UISpecialFrames, "ConverseFrame");

	ConverseFrame:RegisterForDrag("LeftButton");

	Converse_InitHooks();

end

function Converse_InitHooks()
	ConverseFrame:RegisterEvent("VARIABLES_LOADED");
	ConverseFrame:RegisterEvent("GUILD_ROSTER_UPDATE");
	ConverseFrame:RegisterEvent("FRIENDLIST_UPDATE");

	
	-- self could be messy...

	SlashCmdList["CONVERSE"] =
		function(args)
			Converse_SlashCmd(args)
		end

end

function Converse_InitData()
	-- Setup local constants
	RealmName = GetRealmName();
	PlayerName = UnitName("player");

	-- Wheee....
	if (IsInGuild()) then
		GuildRoster();
	end
	GuildName = GetGuildInfo("player");
	if (not GuildName) then
		GuildName = "";
	end

	-- *** Options ***
	if not ConverseState then
		ConverseState = {};
	end
	if not ConverseState[RealmName] then
		ConverseState[RealmName] = {};
	end

	if not ConverseState[RealmName][PlayerName] then
		ConverseState[RealmName][PlayerName] = {};
		-- A few default settings
		ConverseState[RealmName][PlayerName].hidden = 1;
		ConverseState[RealmName][PlayerName].chanblink = 1;
		ConverseState[RealmName][PlayerName].msgblink = 1;
		ConverseState[RealmName][PlayerName].datestamp = 1;
		ConverseState[RealmName][PlayerName].gemfilter = nil;
		ConverseState[RealmName][PlayerName].classcolor = 1;
		
	end

	if not ConverseState[RealmName][PlayerName].fontsize then
		ConverseState[RealmName][PlayerName].fontsize = CONVERSE_DEF_FONTSIZE;
	end
	if not ConverseState[RealmName][PlayerName].lastselected then
		ConverseState[RealmName][PlayerName].lastselected = CONVERSE_DEF_SELECTED;
	end
	if not ConverseState[RealmName][PlayerName].maxlines then
		ConverseState[RealmName][PlayerName].maxlines = CONVERSE_DEF_MAXLINES;
	end
	if not ConverseState[RealmName][PlayerName].ignore then
		ConverseState[RealmName][PlayerName].ignore = {};
		ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_OFFICER] = 1;
	end


	-- *** ConverseHistory ***
	if not ConverseHistory then ConverseHistory = {}; end
	if not ConverseHistory[RealmName] then ConverseHistory[RealmName] = {}; end

	-- Setup the default channels
	-- Say
	if not ConverseHistory[RealmName][CONVERSE_INDEX_SAY] then
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_SAY].text then
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY].text = {};
	end
	-- Party
	if not ConverseHistory[RealmName][CONVERSE_INDEX_PARTY] then
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].text then
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].text = {};
	end
	-- Raid
	if not ConverseHistory[RealmName][CONVERSE_INDEX_RAID] then
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_RAID].text then
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID].text = {};
	end
	-- Instance
	if not ConverseHistory[RealmName][CONVERSE_INDEX_INS] then
		ConverseHistory[RealmName][CONVERSE_INDEX_INS] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_INS].text then
		ConverseHistory[RealmName][CONVERSE_INDEX_INS].text = {};
	end
	-- Guild
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text = {};
	end
	-- Officer
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text = {}; 
	end


	Converse_RefreshChannelsList();

	if not ConverseState[RealmName][PlayerName].multipaste then
		ConverseState[RealmName][PlayerName].multipaste = "";
	end
end

function Converse_HistoryScroll(self, arg1)
	if arg1 > 0 then
		ConverseFrameHistoryScrollFrame:ScrollUp()
		ConverseFrameHistoryScrollFrame:ScrollUp()
	elseif arg1 < 0 and IsAltKeyDown() then
		ConverseFrameHistoryScrollFrame:ScrollToBottom()
	elseif arg1 < 0 then
		ConverseFrameHistoryScrollFrame:ScrollDown()
		ConverseFrameHistoryScrollFrame:ScrollDown()
	end
	lastScroll = GetTime()
	autoScroll = 0
end

function Converse_InitState()

	--	Setup the historyframe
	ConverseFrameHistoryScrollFrame:SetFading(false)
	ConverseFrameHistoryScrollFrame:SetMaxLines(ConverseState[RealmName][PlayerName].maxlines)
	ConverseFrameHistoryScrollFrame:EnableMouseWheel(true);
	ConverseFrameHistoryScrollFrame:SetScript("OnHyperlinkClick", function(self,link,text,button) SetItemRef(link,text,button) end)
	ConverseFrameHistoryScrollFrame:SetScript("OnMouseWheel", Converse_HistoryScroll)

	local ConverseButtonScrollBottom = CreateFrame("Button", "ConverseButtonScrollBottom", ConverseFrame)
	ConverseButtonScrollBottom:SetNormalTexture("Interface/PaperDollInfoFrame/UI-Character-SkillsPageDown-Up")
	ConverseButtonScrollBottom:SetPushedTexture("Interface/PaperDollInfoFrame/UI-Character-SkillsPageDown-Down")
	ConverseButtonScrollBottom:SetHighlightTexture("Interface/Buttons/UI-Common-MouseHilight")
	ConverseButtonScrollBottom:SetWidth(32)
	ConverseButtonScrollBottom:SetHeight(32)
	ConverseButtonScrollBottom:SetPoint("BOTTOMLEFT", ConverseFrameHistoryScrollFrame, "BOTTOMRIGHT", 4, 0)
	ConverseButtonScrollBottom:SetScript("OnClick", function()
		PlaySound(SOUNDKIT.IG_CHAT_BOTTOM)
		ConverseFrameHistoryScrollFrame:ScrollToBottom()
	end)
	
	local ConverseButtonScrollPageDown = CreateFrame("Button", "ConverseButtonScrollPageDown", ConverseFrame)
	ConverseButtonScrollPageDown:SetNormalTexture("Interface/ChatFrame/UI-ChatIcon-ScrollDown-Up")
	ConverseButtonScrollPageDown:SetPushedTexture("Interface/ChatFrame/UI-ChatIcon-ScrollDown-Down")
	ConverseButtonScrollPageDown:SetHighlightTexture("Interface/Buttons/UI-Common-MouseHilight")
	ConverseButtonScrollPageDown:SetWidth(32)
	ConverseButtonScrollPageDown:SetHeight(32)
	ConverseButtonScrollPageDown:SetPoint("BOTTOM", ConverseButtonScrollBottom, "TOP", 1, -5)
	ConverseButtonScrollPageDown:SetScript("OnClick", function()
		PlaySound(SOUNDKIT.IG_CHAT_SCROLL_DOWN)
		ConverseFrameHistoryScrollFrame:PageDown()
	end)
	
	local ConverseButtonScrollDown = CreateFrame("Button", "ConverseButtonScrollDown", ConverseFrame)
	ConverseButtonScrollDown:SetNormalTexture("Interface/ChatFrame/UI-ChatIcon-ScrollDown-Disabled")
	ConverseButtonScrollDown:SetPushedTexture("Interface/ChatFrame/UI-ChatIcon-ScrollDown-Down")
	ConverseButtonScrollDown:SetHighlightTexture("Interface/Buttons/UI-Common-MouseHilight")
	ConverseButtonScrollDown:SetWidth(32)
	ConverseButtonScrollDown:SetHeight(32)
	ConverseButtonScrollDown:SetPoint("BOTTOM", ConverseButtonScrollPageDown, "TOP", 0, -5)
	ConverseButtonScrollDown:SetScript("OnClick", function()
		PlaySound(SOUNDKIT.IG_CHAT_SCROLL_DOWN)
		ConverseFrameHistoryScrollFrame:ScrollDown()
	end)
	ConverseButtonScrollDown:SetScript("OnUpdate", function()
		if ConverseButtonScrollDown:GetButtonState() == "PUSHED" then
			ConverseFrameHistoryScrollFrame:ScrollDown()
		end
	end)
	
	local ConverseButtonScrollUp = CreateFrame("Button", "ConverseButtonScrollUp", ConverseFrame)
	ConverseButtonScrollUp:SetNormalTexture("Interface/ChatFrame/UI-ChatIcon-ScrollUp-Disabled")
	ConverseButtonScrollUp:SetPushedTexture("Interface/ChatFrame/UI-ChatIcon-ScrollUp-Down")
	ConverseButtonScrollUp:SetHighlightTexture("Interface/Buttons/UI-Common-MouseHilight")
	ConverseButtonScrollUp:SetWidth(32)
	ConverseButtonScrollUp:SetHeight(32)
	ConverseButtonScrollUp:SetPoint("BOTTOM", ConverseButtonScrollDown, "TOP", 0, -5)
	ConverseButtonScrollUp:SetScript("OnClick", function()
		PlaySound(SOUNDKIT.IG_CHAT_SCROLL_UP)
		ConverseFrameHistoryScrollFrame:ScrollUp()
	end)
	ConverseButtonScrollUp:SetScript("OnUpdate", function()
		if ConverseButtonScrollUp:GetButtonState() == "PUSHED" then
			ConverseFrameHistoryScrollFrame:ScrollUp()
		end
	end)
	
	local ConverseButtonScrollPageUp = CreateFrame("Button", "ConverseButtonScrollPageUp", ConverseFrame)
	ConverseButtonScrollPageUp:SetNormalTexture("Interface/ChatFrame/UI-ChatIcon-ScrollUp-Up")
	ConverseButtonScrollPageUp:SetPushedTexture("Interface/ChatFrame/UI-ChatIcon-ScrollUp-Down")
	ConverseButtonScrollPageUp:SetHighlightTexture("Interface/Buttons/UI-Common-MouseHilight")
	ConverseButtonScrollPageUp:SetWidth(32)
	ConverseButtonScrollPageUp:SetHeight(32)
	ConverseButtonScrollPageUp:SetPoint("BOTTOM", ConverseButtonScrollUp, "TOP", 0, -5)
	ConverseButtonScrollPageUp:SetScript("OnClick", function()
		PlaySound(SOUNDKIT.IG_CHAT_SCROLL_UP)
		ConverseFrameHistoryScrollFrame:PageUp()
	end)
	
	local ConverseButtonScrollTop = CreateFrame("Button", "ConverseButtonScrollTop", ConverseFrame)
	ConverseButtonScrollTop:SetNormalTexture("Interface/PaperDollInfoFrame/UI-Character-SkillsPageUp-Up")
	ConverseButtonScrollTop:SetPushedTexture("Interface/PaperDollInfoFrame/UI-Character-SkillsPageUp-Down")
	ConverseButtonScrollTop:SetHighlightTexture("Interface/Buttons/UI-Common-MouseHilight")
	ConverseButtonScrollTop:SetWidth(32)
	ConverseButtonScrollTop:SetHeight(32)
	ConverseButtonScrollTop:SetPoint("BOTTOM", ConverseButtonScrollPageUp, "TOP", -1, -5)
	ConverseButtonScrollTop:SetScript("OnClick", function()
		PlaySound(SOUNDKIT.IG_CHAT_BOTTOM)
		ConverseFrameHistoryScrollFrame:ScrollToTop()  
	end)

	
	local i;
	
	--cDateStamp = CONVERSE_DATESTAMP_PRE..date()..CONVERSE_DATESTAMP_POST;
	cDateStamp = string.format(CONVERSE_DATESTAMP_FORMAT, PlayerName, date());

	ConverseItems_Rebuild();

	if (cSelected > cNumItems) then
		-- We have a discrepancy!
		cSelected = 1;
		ConverseState[RealmName][PlayerName].lastselected = cItems[cSelected].key;
	end
	ConverseItems_Update();

	if (cNumItems > 0) and (cSelected > 0) then
		ConverseFrameItemsSelectedText:SetText(cItems[cSelected].name);
		ConverseFrameItemsSelectedText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);

		-- And set up the history window
		ConverseHistory_Fillup(ConverseState[RealmName][PlayerName].lastselected);
	else
		-- BLANK!
		ConverseFrameItemsSelectedText:SetText("");
		ConverseFrameHistoryScrollFrame:Clear();
	end


	-- Font size
	ConverseHistory_SetFontSize(ConverseState[RealmName][PlayerName].fontsize);

	-- Paste frame text
	ConverseFrameMultipasteFrameScrollFrameText:SetText(ConverseState[RealmName][PlayerName].multipaste);

	-- 'Finished' loading.
	cFinishedLoading = 1;
	DEFAULT_CHAT_FRAME:AddMessage(CONVERSE_MSG_LOADED);

	-- Set up text labels and all that fun stuff
	ConverseFrameTitleText:SetText(CONVERSE_HEADER_TITLE.." ("..CONVERSE_VER..") - "..RealmName);
	ConverseFrameIgnoreText:SetText(CONVERSE_HEADER_IGNORE);

	-- Ignore checkbuttons
	for i=1,CONVERSE_MAXCHANNELS do
		_G["ConverseFrameIgnoreCheckButton"..i.."Text"]:SetText(i);
		if ConverseState[RealmName][PlayerName].ignore[i] then
			_G["ConverseFrameIgnoreCheckButton"..i]:SetChecked(1);
		end
	end
	-- Ignore say
	_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_SAY.."Text"]:SetText(CONVERSE_BUTTON_IGNORESAY);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_SAY]:SetChecked(1);
	end
	-- Ignore party
	_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_PARTY.."Text"]:SetText(CONVERSE_BUTTON_IGNOREPARTY);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_PARTY] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_PARTY]:SetChecked(1);
	end
	-- Ignore raid
	_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_RAID.."Text"]:SetText(CONVERSE_BUTTON_IGNORERAID);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_RAID]:SetChecked(1);
	end
	-- Ignore BG
	_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_INS.."Text"]:SetText(CONVERSE_BUTTON_IGNOREINS);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_INS] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_INS]:SetChecked(1);
	end
	-- Ignore guild
	_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_GUILD.."Text"]:SetText(CONVERSE_BUTTON_IGNOREGUILD);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_GUILD] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_GUILD]:SetChecked(1);
	end
	-- Ignore officer
	_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_OFFICER.."Text"]:SetText(CONVERSE_BUTTON_IGNOREOFFICER);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_OFFICER] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_OFFICER]:SetChecked(1);
	end
	-- Ignore whispers
	_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_WHISPER.."Text"]:SetText(CONVERSE_BUTTON_IGNOREWHISPER);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_WHISPER]:SetChecked(1);
	end

	-- Window lock toggle
	if ConverseState[RealmName][PlayerName].locked then
		ConverseFrameLockButton:SetChecked(1);
	else
		ConverseFrameLockButton:SetChecked(0);
	end

	-- And is the frame visible or not
	if ConverseState[RealmName][PlayerName].hidden then
		ConverseFrame:Hide();
	else
		ConverseFrame:Show();
	end
end

function Converse_RefreshChannelsList()
	local i, cnum, cname, nname;
	local key, value;
	local button;
	local schan = {};

	-- Figure out the server channels
	cServerChannels = {};
	schan[1], schan[2], schan[3], schan[4], schan[5], schan[6], schan[7], schan[8], schan[9], schan[10] = EnumerateServerChannels();
	for i = 1,table.getn(schan) do
		-- Channel name starts with letter, but could possible contain digits...
		cname = string.lower(string.sub(schan[i], string.find(schan[i], "%a[%a%d]+")));
		cServerChannels[cname] = GetChannelName(cname);
	end

	-- Make our channels list
	cChannels = {};
	for i=1,CONVERSE_MAXCHANNELS do
		cnum, cname = GetChannelName(i);
		button = _G["ConverseFrameIgnoreCheckButton"..i];
		if cnum and (cnum ~= 0) and (cnum == i) then
			-- Remove everything after the first space (eg. " - Darnassus")
			nname = string.lower(string.sub(cname, string.find(cname,"[^ ]*")));
			-- Ignore the entry if it's the same as cChannelKludge
			if (nname ~= cChannelKludge) then
				cChannels[nname] = cnum;
			else
				-- Remove the history entry
				-- ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..nname] = nil;
			end
			-- Enable the button
			button:Enable();
			button:Show();
		else
			-- Disable the button
			button:SetChecked(0);
			button:Disable();
			button:Hide();
		end
	end

	-- Make sure history entries are there.
	for key, value in pairs(cChannels) do
		key = string.lower(key);
		if not ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key] then
			ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key] = {}
		end
		if not ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key].text then
			ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key].text = {};
		end
	end
end


--=== Console interface ====================================================--
function Converse_SlashCmd(args)
	local i;

	if (args) then
		local opts = string.lower(args);
		if (opts == "") then
			Converse_ShowHelp();
		elseif (opts == CONVERSE_SLASH_HIDE) then
			ConverseFrame:Hide();
			ConverseState[RealmName][PlayerName].hidden = 1;
			PlaySound(SOUNDKIT.IG_QUEST_LIST_OPEN);
		elseif (opts == CONVERSE_SLASH_SHOW) then
			ConverseFrame:Show();
			ConverseState[RealmName][PlayerName].hidden = nil;
			PlaySound(SOUNDKIT.IG_QUEST_LIST_CLOSE);
		elseif (opts == CONVERSE_SLASH_LOCK) then
			ConverseState[RealmName][PlayerName].locked = 1;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_LOCKED);
			ConverseFrameLockButton:SetChecked(1);
		elseif (opts == CONVERSE_SLASH_UNLOCK) then
			ConverseState[RealmName][PlayerName].locked = nil;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_UNLOCKED);
			ConverseFrameLockButton:SetChecked(0);
		elseif (opts == CONVERSE_SLASH_RESET) then
			ConverseFrame:ClearAllPoints();
			ConverseFrame:SetPoint("TOPRIGHT", "UIParent", "TOPRIGHT", -16, -104);
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_RESET);
		elseif (string.find(opts, CONVERSE_SLASH_FONTSIZE.." ") == 1) then
			i = tonumber(string.sub(opts, string.len(CONVERSE_SLASH_FONTSIZE) + 2, string.len(opts)));
			if (i >= CONVERSE_FONTMIN) and (i <= CONVERSE_FONTMAX) then
				ConverseHistory_SetFontSize(i);
				ConverseState[RealmName][PlayerName].fontsize = i;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_FONTSIZE..i);
			else
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_ERR_FONTSIZE..i);
			end
		elseif (string.find(opts, CONVERSE_SLASH_MAXLINES.." ") == 1) then
			i = tonumber(string.sub(opts, string.len(CONVERSE_SLASH_MAXLINES) + 2, string.len(opts)));
			if (i == 0) then
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_ERR_MAXLINES..i);
				--ConverseState[RealmName][PlayerName].maxlines = 0;
				-- Infinite. BEWARE!
			elseif (i >= CONVERSE_LINESMIN) and (i <= CONVERSE_LINESMAX) then
				ConverseState[RealmName][PlayerName].maxlines = i;
				ConverseFrameHistoryScrollFrame:SetMaxLines(ConverseState[RealmName][PlayerName].maxlines)
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_MAXLINES..i);
			else
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_ERR_MAXLINES..i);
			end
		-- Blinkinlights!
		elseif (opts == CONVERSE_SLASH_BLINK) then
			if not ConverseState[RealmName][PlayerName].blink then
				ConverseState[RealmName][PlayerName].blink = 1;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_BLINK..CONVERSE_MSG_ENABLED);
			else
				ConverseState[RealmName][PlayerName].blink = nil;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_BLINK..CONVERSE_MSG_DISABLED);
			end
			ConverseItems_Rebuild();
			ConverseItems_Update();
		elseif (opts == CONVERSE_SLASH_MSGBLINK) then
			if not ConverseState[RealmName][PlayerName].msgblink then
				ConverseState[RealmName][PlayerName].msgblink = 1;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_MSGBLINK..CONVERSE_MSG_ENABLED);
			else
				ConverseState[RealmName][PlayerName].msgblink = nil;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_MSGBLINK..CONVERSE_MSG_DISABLED);
			end
			ConverseItems_Rebuild();
			ConverseItems_Update();
		elseif (opts == CONVERSE_SLASH_CHANBLINK) then
			if not ConverseState[RealmName][PlayerName].chanblink then
				ConverseState[RealmName][PlayerName].chanblink = 1;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_CHANBLINK..CONVERSE_MSG_ENABLED);
			else
				ConverseState[RealmName][PlayerName].chanblink = nil;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_CHANBLINK..CONVERSE_MSG_DISABLED);
			end
			ConverseItems_Rebuild();
			ConverseItems_Update();
		-- Class color
		elseif (opts == CONVERSE_SLASH_CLASSCOLOR) then
			if not ConverseState[RealmName][PlayerName].classcolor then
				ConverseState[RealmName][PlayerName].classcolor = 1;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLASSCOLOR..CONVERSE_MSG_ENABLED);
			else
				ConverseState[RealmName][PlayerName].classcolor = nil;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLASSCOLOR..CONVERSE_MSG_DISABLED);
			end
			ConverseItems_Rebuild();
			ConverseItems_Update();

		-- Clear
		elseif (opts == CONVERSE_SLASH_CLEARALL) then
			Converse_Clear(CONVERSE_CLEARALL);
		elseif (opts == CONVERSE_SLASH_CLEARCHAT) then
			Converse_Clear(CONVERSE_CLEARCHAT);
		elseif (opts == CONVERSE_SLASH_CLEARCHAN) then
			Converse_Clear(CONVERSE_CLEARCHAN);
		elseif (opts == CONVERSE_SLASH_CLEARMSG) then
			Converse_Clear(CONVERSE_CLEARMSG);
		elseif (opts == CONVERSE_SLASH_CLEARRAND) then
			Converse_Clear(CONVERSE_CLEARRAND);
		-- Date stamping
		elseif (opts == CONVERSE_SLASH_DATE) then
			ConverseState[RealmName][PlayerName].datestamp = 1;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_DATE..CONVERSE_MSG_ENABLED);
		elseif (opts == CONVERSE_SLASH_NODATE) then
			ConverseState[RealmName][PlayerName].datestamp = nil;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_DATE..CONVERSE_MSG_DISABLED);
		-- GEM filter
		elseif (opts == CONVERSE_SLASH_GEMFILTER) then
			ConverseState[RealmName][PlayerName].gemfilter = 1;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_GEMFILTER..CONVERSE_MSG_ENABLED);
		elseif (opts == CONVERSE_SLASH_NOGEMFILTER) then
			ConverseState[RealmName][PlayerName].gemfilter = nil;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_GEMFILTER..CONVERSE_MSG_DISABLED);
		-- EOL
		elseif (opts == CONVERSE_SLASH_STATUS) then
			Converse_Status();
		else
			Converse_ShowHelp();
		end
	else
		Converse_ShowHelp();
	end
end

function Converse_Status()
	cMessage(CONVERSE_MSG_STATUSHEADER);
	if ConverseState[RealmName][PlayerName].locked then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_LOCKED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_UNLOCKED);
	end
	cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_FONTSIZE..ConverseState[RealmName][PlayerName].fontsize);
	cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_MAXLINES..ConverseState[RealmName][PlayerName].maxlines);
	if ConverseState[RealmName][PlayerName].blink then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_BLINK..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_BLINK..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].chanblink then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_CHANBLINK..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_CHANBLINK..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].msgblink then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_MSGBLINK..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_MSGBLINK..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].classcolor then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_CLASSCOLOR..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_CLASSCOLOR..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].datestamp then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_DATE..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_DATE..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].gemfilter then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_GEMFILTER..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_GEMFILTER..CONVERSE_MSG_DISABLED);
	end
end

function Converse_ShowHelp()
	local i;
		for i=1, table.getn(CONVERSE_HELPTEXT) do
		cMessage(CONVERSE_HELPTEXT[i]);
	end
end

function Converse_Clear(cleartype)
	if (cleartype == CONVERSE_CLEARALL) then
		ConverseHistory[RealmName] = {}
		Converse_InitData();
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARALL..RealmName);
		ConverseItems_Rebuild();
		ConverseItems_Update();
		ConverseFrameHistoryScrollFrame:Clear();
		return;
	elseif (cleartype == CONVERSE_CLEARCHAT) then
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY] = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY].text = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY] = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].text = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID] = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID].text = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_INS] = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_INS].text = {};
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] = {};
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text = {};
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] = {};
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text = {};
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARCHAT..RealmName);
	elseif (cleartype == CONVERSE_CLEARCHAN) then
		local key, value;
		for key, value in pairs(ConverseHistory[RealmName]) do
			pfx = string.sub(key,1,2);
			cname = string.sub(key, 3);
			lname = string.lower(cname);
			if (pfx == CONVERSE_PREFIX_CHANNEL) then
				if cChannels[lname] then
					-- Numbered channel
					ConverseHistory[RealmName][key] = {};
					ConverseHistory[RealmName][key].text = {};
				else
					-- Unnumbered channel
					ConverseHistory[RealmName][key] = nil;
				end
			end
		end
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARCHAN..RealmName);
	elseif (cleartype == CONVERSE_CLEARMSG) then
		-- Messages only
		local key, value;
		for key, value in pairs(ConverseHistory[RealmName]) do
			pfx = string.sub(key,1,2);
			cname = string.sub(key, 3);
			if (pfx == CONVERSE_PREFIX_MSG) then
				ConverseHistory[RealmName][key] = nil;
			end
		end
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARMSG..RealmName);
	elseif (cleartype == CONVERSE_CLEARRAND) then
		local lfriends = {};
		for i=1,GetNumFriends() do
			local fname, flevel, fclass, farea, fconn, fstat = GetFriendInfo(i);
			if fname then
				if fconn then
					lfriends[fname] = 1;
				else
					lfriends[fname] = 0;
				end
			end
		end
		-- Non friend messages only
		local key, value;
		for key, value in pairs(ConverseHistory[RealmName]) do
			pfx = string.sub(key,1,2);
			cname = string.sub(key, 3);
			if (pfx == CONVERSE_PREFIX_MSG) and not lfriends[cname] then
				ConverseHistory[RealmName][key] = nil;
			end
		end
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARRAND..RealmName);
	else
		return;
		-- Huh?!??
	end

	-- Update;
	ConverseItems_Rebuild();
	ConverseItems_Update();
	if (cNumItems > 0) and (cSelected > 0) then
		ConverseHistory_Fillup(cItems[cSelected].key);
	else
		ConverseFrameHistoryScrollFrame:Clear();
	end
end

--=== The guts =============================================================--

function Converse_OnEvent(event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
	local cname, lname;
	local pname;
	
	if autoScroll == 0 then
		if (lastScroll - GetTime()) < -45 then
			autoScroll = 1;
			ConverseFrameHistoryScrollFrame:ScrollToBottom();
		end
	end
	
	cChannelKludge = "";

	if (event == "VARIABLES_LOADED") then
		Converse_InitData();
		Converse_InitState();
		hooksecurefunc("ChatFrame_OnEvent",Converse_ParseChatMessage);
	elseif (event == "GUILD_ROSTER_UPDATE") then
		if cFinishedLoading then
			-- And more kludginess
			Converse_GuildKludge()

			-- Refresh
			ConverseItems_Rebuild();
			ConverseItems_Update();
		end
	elseif (event == "FRIENDLIST_UPDATE") then
		if cFinishedLoading then
			ConverseItems_Rebuild();
			ConverseItems_Update();
		end
	end
end

function Converse_GuildKludge()
	GuildName = GetGuildInfo("player");
	if (not GuildName) then
		GuildName = "";
	end
	-- Guild
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text = {};
	end
	-- Officer
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text = {}; 
	end
end

function Converse_OnShow()
	ConverseState[RealmName][PlayerName].hidden = nil;

	GuildName = GetGuildInfo("player");
	if (not GuildName) then
		GuildName = "";
	end

	-- Button States
	for i=1,CONVERSE_MAXCHANNELS do
		if ConverseState[RealmName][PlayerName].ignore[i] then
			_G["ConverseFrameIgnoreCheckButton"..i]:SetChecked(1);
		end
	end
	-- And the other 6...
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_SAY] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_SAY]:SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_PARTY] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_PARTY]:SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_RAID] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_RAID]:SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_INS] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_INS]:SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_GUILD] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_GUILD]:SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_OFFICER] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_OFFICER]:SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_WHISPER] then
		_G["ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_WHISPER]:SetChecked(1);
	end

	ConverseItems_Rebuild();
	ConverseItems_Update();

	-- Just for fun
	SetPortraitTexture(ConverseFrameIconTexture, "player");


end

function Converse_OnHide()
	ConverseState[RealmName][PlayerName].hidden = 1;
end

function Converse_ParseChatMessage(self, event, text, author, language, channumname, target, flags, zoneid, channum, channame, _, _, guid)
	-- Don't you love those args?

	local key;
	local cname;
	local title;
	local oneline = nil ;
	local timestamp;
	local i;
	local authcolor = ""
	local coloredname

	if guid and guid ~= "" then
		local _, englishClass = GetPlayerInfoByGUID(guid)
		if englishClass then
			local classColorTable = RAID_CLASS_COLORS[englishClass];
			if classColorTable then
				authcolor = string.format("|cff%02x%02x%02x", classColorTable.r * 255, classColorTable.g * 255, classColorTable.b * 255)
			end
		end
	end


	-- By default, self will be...
	if author and type(author) == "string" then
		if authcolor ~= "" and ConverseState[RealmName][PlayerName].classcolor then
			coloredname = authcolor.."|Hplayer:"..author.."|h["..author.."]|h|r";
			title = CONVERSE_NAME_PRE..authcolor.."|Hplayer:"..author.."|h["..author.."]|h|r"..CONVERSE_NAME_POST;
		else
			title = CONVERSE_NAME_PRE.."|Hplayer:"..author.."|h["..author.."]|h"..CONVERSE_NAME_POST; 
			coloredname = CONVERSE_NAME_PRE.."|Hplayer:"..author.."|h["..author.."]|h"..CONVERSE_NAME_POST;
		end
	end
	
	if (event == "CHAT_MSG_WHISPER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		key = CONVERSE_PREFIX_MSG..author;
	elseif (event == "CHAT_MSG_WHISPER_INFORM") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		key = CONVERSE_PREFIX_MSG..author;
		--title = CONVERSE_NAME_PRE..PlayerName..CONVERSE_NAME_REPLY..author..CONVERSE_NAME_POST;
		title = PlayerName..CONVERSE_NAME_REPLY..CONVERSE_NAME_PRE..coloredname..CONVERSE_NAME_POST;
	elseif (event == "CHAT_MSG_AFK") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		key = CONVERSE_PREFIX_MSG..author;
		title = coloredname..CONVERSE_NAME_AFK;
	elseif (event == "CHAT_MSG_DND") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		key = CONVERSE_PREFIX_MSG..author;
		title = coloredname..CONVERSE_NAME_DND;
	elseif (event == "CHAT_MSG_GUILD") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_GUILD] then return; end
		key = CONVERSE_PREFIX_GUILD..GuildName;
	elseif (event == "CHAT_MSG_OFFICER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_OFFICER] then return; end
		key = CONVERSE_PREFIX_OFFICER..GuildName;
	elseif (event == "CHAT_MSG_PARTY") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_PARTY] then return; end
		key = CONVERSE_INDEX_PARTY;
	elseif (event == "CHAT_MSG_PARTY_LEADER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_PARTY] then return; end
		key = CONVERSE_INDEX_PARTY;
		title = CONVERSE_NAME_RLEADPRE..coloredname..CONVERSE_NAME_POST;
	-- Say, yell, and all that fun stuff
	elseif (event == "CHAT_MSG_SAY") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = coloredname..CONVERSE_NAME_SAY;
	elseif (event == "CHAT_MSG_YELL") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = coloredname..CONVERSE_NAME_YELL;
	elseif (event == "CHAT_MSG_EMOTE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = coloredname.." ";
	elseif (event == "CHAT_MSG_TEXT_EMOTE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = "";
	-- Monsters!
	elseif (event == "CHAT_MSG_MONSTER_SAY") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author..CONVERSE_NAME_SAY;
	elseif (event == "CHAT_MSG_MONSTER_YELL") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author..CONVERSE_NAME_YELL;
	elseif (event == "CHAT_MSG_MONSTER_EMOTE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author.." ";
	-- Raid, Raid leader, Raid warning
	elseif (event == "CHAT_MSG_RAID") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then return; end
		key = CONVERSE_INDEX_RAID;
	elseif (event == "CHAT_MSG_RAID_LEADER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then return; end
		key = CONVERSE_INDEX_RAID;
		title = CONVERSE_NAME_RLEADPRE..coloredname..CONVERSE_NAME_POST;
	elseif (event == "CHAT_MSG_RAID_WARNING") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then return; end
		key = CONVERSE_INDEX_RAID;
		title = CONVERSE_NAME_RWARNPRE..coloredname..CONVERSE_NAME_POST;
	-- Instance and Instance Leader and in-BG messages
	elseif (event == "CHAT_MSG_INSTANCE_CHAT") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_INS] then return; end
		key = CONVERSE_INDEX_INS;
	elseif (event == "CHAT_MSG_INSTANCE_CHAT_LEADER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_INS] then return; end
		key = CONVERSE_INDEX_INS;
		title = CONVERSE_NAME_BLEADPRE..coloredname..CONVERSE_NAME_POST;
	elseif (event == "CHAT_MSG_BG_SYSTEM_ALLIANCE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_INS] then return; end
		key = CONVERSE_INDEX_INS;
		title = cChatColourString("BG_SYSTEM_ALLIANCE");
		text = text.."|r";
	elseif (event == "CHAT_MSG_BG_SYSTEM_HORDE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_INS] then return; end
		key = CONVERSE_INDEX_INS;
		title = cChatColourString("BG_SYSTEM_HORDE");
		text = text.."|r";
	elseif (event == "CHAT_MSG_BG_SYSTEM_NEUTRAL") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_INS] then return; end
		key = CONVERSE_INDEX_INS;
		title = cChatColourString("BG_SYSTEM_NEUTRAL");
		text = text.."|r";
	-- Channel
	elseif (event == "CHAT_MSG_CHANNEL") then
		if ConverseState[RealmName][PlayerName].ignore[channum] then return; end
		cname = string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		key = CONVERSE_PREFIX_CHANNEL..cname;
		title = CONVERSE_NAME_PRE..coloredname..CONVERSE_NAME_POST;
	elseif (event == "CHAT_MSG_CHANNEL_JOIN") then
		--key = CONVERSE_PREFIX_CHANNEL..string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		cname = string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		key = CONVERSE_PREFIX_CHANNEL..cname;
		---
		----
		if ConverseState[RealmName][PlayerName].ignore[channum] then return; end
		title = coloredname..CONVERSE_NAME_JOIN;
		text = "";
	elseif (event == "CHAT_MSG_CHANNEL_LEAVE") then
		--key = CONVERSE_PREFIX_CHANNEL..string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		cname = string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		key = CONVERSE_PREFIX_CHANNEL..cname;
		---
		----
		if ConverseState[RealmName][PlayerName].ignore[channum] then return; end
		title = coloredname..CONVERSE_NAME_LEAVE;
		text = "";
	elseif (event == "GUILD_MOTD") then
		if string.len(text) > 0 then
			key = CONVERSE_PREFIX_GUILD..GuildName;
			title = "";
			text = string.format(TEXT(GUILD_MOTD_TEMPLATE), text);
		else
			return;
		end
	else
		-- SHOULD NOT GET self!
		return;
	end
	-- Filter out GEM messages
	if ConverseState[RealmName][PlayerName].gemfilter then
		if string.find(text,"<GEM%d+.%d+") == 1 then
			return;
		end
	end
	
	timestamp = string.sub(date(),10,17);


	-- Does self item exist in the history?
	if not ConverseHistory[RealmName][key] then
		ConverseHistory[RealmName][key] = {};
		if ConverseState[RealmName][PlayerName].datestamp then
			local dateline = cChatColourString(string.sub(event,10), true)..cDateStamp.."|r";
			ConverseHistory[RealmName][key].text = {dateline};
			ConverseHistory[RealmName][key].datestamp = cDateStamp;
		else
			ConverseHistory[RealmName][key].text = {};
		end
	else
		-- Check if we're adding a datestamp, or replacing a datestamp
		if ConverseState[RealmName][PlayerName].datestamp and
		   ((not ConverseHistory[RealmName][key].datestamp) or (ConverseHistory[RealmName][key].datestamp ~= cDateStamp)) then
			local dateline = "";
			if (string.sub(key,1,2) == CONVERSE_PREFIX_CHANNEL) then
				-- Account for colour numbers
				if cname and cChannels[cname] then
					dateline = cChatColourString("CHANNEL"..cChannels[cname], true);
				end
			else
				dateline = cChatColourString(string.sub(event,10), true);
			end
			dateline = dateline..cDateStamp.."|r";

			ConverseHistory[RealmName][key].datestamp = cDateStamp;
			table.insert(ConverseHistory[RealmName][key].text, dateline)

			if (key == ConverseState[RealmName][PlayerName].lastselected) then
				ConverseFrameHistoryScrollFrame:AddMessage(dateline ) 

				if (ConverseState[RealmName][PlayerName].maxlines > 0) then
					ConverseFrameLinesText:SetText(ConverseFrameHistoryScrollFrame:GetNumMessages().."/"..ConverseState[RealmName][PlayerName].maxlines);
				else
					ConverseFrameLinesText:SetText(ConverseFrameHistoryScrollFrame:GetNumMessages()..CONVERSE_HEADER_LINES);
				end
			end		
		end
	end


	local Mlines = ConverseState[RealmName][PlayerName].maxlines
		if #ConverseHistory[RealmName][key].text > ( Mlines + 50) then
			while #ConverseHistory[RealmName][key].text > Mlines do
				table.remove(ConverseHistory[RealmName][key].text, 1);
			end
		end
	-- Build the line
	
	oneline = title..text;

	-- Colour Mangling
	if (event == "CHAT_MSG_EMOTE") or
	   (event == "CHAT_MSG_TEXT_EMOTE") or
	   (event == "CHAT_MSG_YELL") or
	   (event == "CHAT_MSG_MONSTER_SAY") or
	   (event == "CHAT_MSG_MONSTER_EMOTE") or
	   (event == "CHAT_MSG_MONSTER_YELL") then
		local chattype = string.sub(event, 10);
		local linecolour = cChatColourString(chattype);
		oneline = linecolour..oneline;

		-- Look for |h|r's in the string, and append the colour again, otherwise it'll revert to the 'Say' colour
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);

		-- And the trailing colour return
		oneline = oneline.."|r";
	elseif (event == "CHAT_MSG_CHANNEL_JOIN") or (event == "CHAT_MSG_CHANNEL_LEAVE") then
		oneline = CONVERSE_COLOUR_GREY..oneline.."|r";
	elseif (event == "CHAT_MSG_RAID_LEADER") then
		local linecolour = cChatColourString("RAID_LEADER");
		oneline = linecolour..oneline;
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);
		oneline = oneline.."|r";
	elseif (event == "CHAT_MSG_RAID_WARNING") then
		local linecolour = cChatColourString("RAID_WARNING");
		oneline = linecolour..oneline;
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);
		oneline = oneline.."|r";
	elseif (event == "CHAT_MSG_INSTANCE_LEADER") then
		local linecolour = cChatColourString("INSTANCE_LEADER");
		oneline = linecolour..oneline;
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);
		oneline = oneline.."|r";
	end

	-- Timestamp
	oneline = timestamp.." "..oneline;

	-- MWBIGChat hook
	if MWBIGChat then
		MWBIGChat(event, text, author, language, channumname, target, flags, zoneid, channum, channame)
	end

	-- DONE!
	
	-- debug
	
	-- Add it to the history
	if oneline ~= ConverseHistory[RealmName][key].text[#ConverseHistory[RealmName][key].text] then
		table.insert(ConverseHistory[RealmName][key].text, oneline)
	

	-- DEBUG: cMessage("CM | key = "..key.." | oneline = "..oneline);

	-- if current item, update history
		if (key == ConverseState[RealmName][PlayerName].lastselected) then
			ConverseFrameHistoryScrollFrame:AddMessage(oneline, cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b  ) 

			if (ConverseState[RealmName][PlayerName].maxlines > 0) then
				ConverseFrameLinesText:SetText(ConverseFrameHistoryScrollFrame:GetNumMessages().."/"..ConverseState[RealmName][PlayerName].maxlines);
			else
				ConverseFrameLinesText:SetText(ConverseFrameHistoryScrollFrame:GetNumMessages()..CONVERSE_HEADER_LINES);
			end
		end
	end
	
	-- Required for flashing...
	-- No activity on channel joins/leaves
	if (event ~= "CHAT_MSG_CHANNEL_JOIN") and
	   (event ~= "CHAT_MSG_CHANNEL_LEAVE") and
	   (event ~= "CHAT_MSG_WHISPER_INFORM") then
	   	if not ConverseHistory[RealmName][key].activity then
			ConverseHistory[RealmName][key].activity = 1;
		else
			ConverseHistory[RealmName][key].activity = ConverseHistory[RealmName][key].activity + 1;
		end
		ConverseHistory[RealmName][key].lastdate = date();
	end

	ConverseItems_Rebuild();

	ConverseItems_Update();
end

--=== Stupid FauxScrollBars! ===============================================--

function ConverseItems_Rebuild()
	local key, value;
	local i, j;
	local nametable = {};
	local pfx, cname, lname;

	-- KLUDGE
	Converse_GuildKludge();

	-- Friends experimental
	local lfriends = {};
	for i=1,GetNumFriends() do
		local fname, flevel, fclass, farea, fconn, fstat = GetFriendInfo(i);
		if not fname then
			--DEFAULT_CHAT_FRAME:AddMessage("NO FNAME i="..i);
		else
			if fconn then
				lfriends[fname] = 1;
			else
				lfriends[fname] = 0;
			end
		end
	end

	-- Default 4
	i = 1;
	cItems = {};

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then
		cItems[i]		= {};
		cItems[i].key 		= CONVERSE_INDEX_SAY;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["SAY"].r;
		cItems[i].colour.g	= ChatTypeInfo["SAY"].g;
		cItems[i].colour.b	= ChatTypeInfo["SAY"].b;
		cItems[i].name		= CONVERSE_LABEL_SAY;
		if ConverseHistory[RealmName][CONVERSE_INDEX_SAY].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_PARTY] then
		cItems[i]		= {};
		cItems[i].key 		= CONVERSE_INDEX_PARTY;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["PARTY"].r;
		cItems[i].colour.g	= ChatTypeInfo["PARTY"].g;
		cItems[i].colour.b	= ChatTypeInfo["PARTY"].b;
		cItems[i].name		= CONVERSE_LABEL_PARTY;
		if ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then
		cItems[i]		= {};
		cItems[i].key		= CONVERSE_INDEX_RAID;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["RAID"].r;
		cItems[i].colour.g	= ChatTypeInfo["RAID"].g;
		cItems[i].colour.b	= ChatTypeInfo["RAID"].b;
		cItems[i].name		= CONVERSE_LABEL_RAID;
		if ConverseHistory[RealmName][CONVERSE_INDEX_RAID].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_INS] then
		cItems[i]		= {};
		cItems[i].key		= CONVERSE_INDEX_INS;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["INSTANCE_CHAT"].r;
		cItems[i].colour.g	= ChatTypeInfo["INSTANCE_CHAT"].g;
		cItems[i].colour.b	= ChatTypeInfo["INSTANCE_CHAT"].b;
		cItems[i].name		= CONVERSE_LABEL_INS;
		if ConverseHistory[RealmName][CONVERSE_INDEX_INS].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_GUILD] then
		cItems[i]		= {};
		cItems[i].key		= CONVERSE_PREFIX_GUILD..GuildName;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["GUILD"].r;
		cItems[i].colour.g	= ChatTypeInfo["GUILD"].g;
		cItems[i].colour.b	= ChatTypeInfo["GUILD"].b;
		cItems[i].name		= CONVERSE_LABEL_GUILD.." <"..GuildName..">";
		if ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_OFFICER] then
		cItems[i]		= {};
		cItems[i].key		= CONVERSE_PREFIX_OFFICER..GuildName;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["OFFICER"].r;
		cItems[i].colour.g	= ChatTypeInfo["OFFICER"].g;
		cItems[i].colour.b	= ChatTypeInfo["OFFICER"].b;
		cItems[i].name		= CONVERSE_LABEL_OFFICER.." <"..GuildName..">";
		if ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	-- Channels and Private messages
	Converse_RefreshChannelsList();
	for key, value in pairs(ConverseHistory[RealmName]) do
		pfx = string.sub(key,1,2);
		cname = string.sub(key, 3);
		lname = string.lower(cname);
		if (pfx == CONVERSE_PREFIX_MSG) then
			-- Private Message: "m/PlayerName"
		   	-- Assuming blizzard doesn't allow 1- or 0-letter names in the future...
			if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then
				if lfriends[cname] == 1 then
					-- Friends we substitue "k/" in front so they end up sorted ahead of other messages "m/"
					table.insert(nametable, CONVERSE_PREFIX_MSGFRIEND..cname);
				elseif lfriends[cname] == 0 then
					-- "l/" for offline friends next
					table.insert(nametable, CONVERSE_PREFIX_MSGFRIENDOFFLINE..cname);
				else
					table.insert(nametable, key);
				end
			end
		elseif (pfx == CONVERSE_PREFIX_CHANNEL) then
			-- Channels: "c/##/Channelname"
			-- We tack on the number so we can sort by number
			if cChannels[lname] then
				if not ConverseState[RealmName][PlayerName].ignore[cChannels[lname]] then
					table.insert(nametable, CONVERSE_PREFIX_CHANNEL..string.format("%02d", cChannels[lname])..lname);
				end
			else
				-- Unnumbered channels (ie. currently not listening to) come last.
				table.insert(nametable, CONVERSE_PREFIX_CHANNEL.."99"..lname);
			end
		end
	end
	table.sort(nametable);

	for j=1,table.getn(nametable) do
		--cMessage(j.." = '"..nametable[j].."'");
		cItems[i] = {};
		pfx = string.sub(nametable[j],1,2)
		if pfx == CONVERSE_PREFIX_CHANNEL then
			-- Remove the extraneous number which was there only for sorting.
			cname = string.sub(nametable[j],5);
			lname = string.lower(cname);
			cItems[i].key		= CONVERSE_PREFIX_CHANNEL..lname;
			cItems[i].colour	= {};
			if cChannels[lname] then
				local _, fullname = GetChannelName(cChannels[lname]);
				cItems[i].name	= cChannels[lname]..". "..fullname;
				-- Experimental:
				cItems[i].colour["r"]	= ChatTypeInfo["CHANNEL"..cChannels[lname]].r;
				cItems[i].colour["g"]	= ChatTypeInfo["CHANNEL"..cChannels[lname]].g; 
				cItems[i].colour["b"]	= ChatTypeInfo["CHANNEL"..cChannels[lname]].b; 
			else
				cItems[i].name	= lname;
				cItems[i].colour["r"]	= CONVERSE_COLOUR_CHANNEL["r"];
				cItems[i].colour["g"]	= CONVERSE_COLOUR_CHANNEL["g"];
				cItems[i].colour["b"]	= CONVERSE_COLOUR_CHANNEL["b"];
			end
		else
			-- Private messages
			cname = string.sub(nametable[j], 3);
			-- Substitute "m/" back for "l/" so we get the right index
			cItems[i].key		= nametable[j];
			if (pfx == CONVERSE_PREFIX_MSGFRIEND) or (pfx == CONVERSE_PREFIX_MSGFRIENDOFFLINE) then
				cItems[i].key	= CONVERSE_PREFIX_MSG..cname;
			end
			cItems[i].colour	= {};
			cItems[i].colour.r	= ChatTypeInfo["WHISPER"].r;
			cItems[i].colour.g	= ChatTypeInfo["WHISPER"].g;
			cItems[i].colour.b	= ChatTypeInfo["WHISPER"].b;
			if lfriends[cname] == 1 then	
				cItems[i].name		= cname.." "..CONVERSE_NAME_ONLINE;
			elseif lfriends[cname] == 0 then	
				cItems[i].name		= cname.." "..CONVERSE_NAME_OFFLINE;
			else
				cItems[i].name		= cname;
			end
		end

		-- Activity tag
		if ConverseHistory[RealmName][cItems[i].key].activity then
			cItems[i].activity = 1;
		else
			cItems[i].activity = nil;
		end

		i = i + 1;
	end

	cNumItems = (i - 1);

	cSelected = -1;
	for i=1,table.getn(cItems) do
		if (cItems[i].key == ConverseState[RealmName][PlayerName].lastselected) then
			cSelected = i;
		end
	end

	if (cSelected == -1) and (cNumItems > 0) then
		-- Okay, we didn't find it. Got cleared or ignored
		cSelected = 1;
		ConverseState[RealmName][PlayerName].lastselected = cItems[cSelected].key;
	end
end

-- Items buttons
function ConverseItems_OnClick(self, button)
	local id = self:GetID();
	button = _G["ConverseFrameItemsButton"..id];

	cSelected = button.listindex;

	if not cItems[cSelected] then
		-- Uh oh, we have a stray item that has no name...
		ConverseItems_Rebuild();
		ConverseItems_Update();
		return;
	end

	ConverseState[RealmName][PlayerName].lastselected = cItems[cSelected].key;

	-- Turn off flashing
	ConverseHistory[RealmName][cItems[cSelected].key].activity = nil;
	cItems[cSelected].activity = nil;

	-- Change history to match
	ConverseHistory_Fillup(cItems[cSelected].key);

	-- Set the 'selected' text.
	ConverseFrameItemsSelectedText:SetText(cItems[cSelected].name);
	ConverseFrameItemsSelectedText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);
	ConverseItems_Update();

end

function ConverseItems_Update()
	local index;
	local offset = FauxScrollFrame_GetOffset(ConverseFrameItemsScrollFrame);
	local button, buttontext;

	for i=1, CONVERSE_ITEMS_NUMSHOWN do
		index = i + offset;
		button = _G["ConverseFrameItemsButton"..i];
		button.listindex = index;

		buttontext = _G["ConverseFrameItemsButton"..i.."Text"];
		if (index <= cNumItems) then
			buttontext:SetText(" ".. cItems[index].name);
			buttontext:SetTextColor(cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b);
			if (cItems[index].activity) then
				if ConverseState[RealmName][PlayerName].blink or
				   (ConverseState[RealmName][PlayerName].chanblink and string.sub(cItems[index].key,1,2) == CONVERSE_PREFIX_CHANNEL) or
				   (ConverseState[RealmName][PlayerName].msgblink and string.sub(cItems[index].key,1,2) == CONVERSE_PREFIX_MSG) then
					if (not button.animFade:IsPlaying()) then
						button.animFade:Play();
					end
				else
					button.animFade:Stop();
				end
			else
				button.animFade:Stop();
			end
			button:Show();
		else
			button:UnlockHighlight();
			buttontext:SetText("");
			button.animFade:Stop();
			button:Hide();
		end

		if (cNumItems > 0) and (index == cSelected) then
			cItems[index].activity = nil;
			ConverseHistory[RealmName][cItems[index].key].activity = nil;

			button:LockHighlight();
			button.animFade:Stop();
			button:Show();
		else
			button:UnlockHighlight();
		end
	end

	FauxScrollFrame_Update(ConverseFrameItemsScrollFrame, cNumItems, CONVERSE_ITEMS_NUMSHOWN, CONVERSE_ITEMS_HEIGHT);

	if (cNumItems > 0) and (cSelected > 0) then
		ConverseFrameItemsSelectedText:SetText(cItems[cSelected].name);
		ConverseFrameItemsSelectedText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);
	else
		ConverseFrameItemsSelectedText:SetText("");
	end

	-- KLUDGE: Since I don't have a graphical divider other than the scrollbar
	ConverseFrameItemsScrollFrame:Show();

end

function ConverseItems_OnEnter(item)
	local index	= item:GetID();

	if not cItems[index] then
		-- Uh oh, we have a stray item that has no name...
		ConverseItems_Rebuild();
		ConverseItems_Update();
		return;
	end

	local key	= cItems[index].key;
	local prefix	= string.sub(cItems[index].key,1,2);
	local cname	= string.sub(cItems[index].key,3);
	local lname	= string.lower(cname);
	local newlines	= 0;

	local numraid	= GetNumGroupMembers();
	local numparty	= GetNumSubgroupMembers();	-- Nice, it's 0-4!

	GameTooltip:SetOwner(_G["ConverseFrameItemsButton"..index], "ANCHOR_BOTTOMRIGHT");
	if (prefix == CONVERSE_PREFIX_MSG) then
		-- Show location of friend
		-- FIXME: I should really have a global friends structure instead of redoing it every time...
		local locfriends = {};
		local i;
		for i=1,GetNumFriends() do
			local fname, flevel, fclass, farea, fconn, fstat = GetFriendInfo(i);
			if fname and fconn then
				if (fstat ~= "") then
					locfriends[fname] = farea.." "..fstat;
				else
					locfriends[fname] = farea;
				end
			end
		end
		if locfriends[cname] then
			GameTooltip:AddDoubleLine(cItems[index].name,
			                          locfriends[cname],
			                          cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b,
						  CONVERSE_COLOUR_TOOLTIP_LOCATION.r, CONVERSE_COLOUR_TOOLTIP_LOCATION.g, CONVERSE_COLOUR_TOOLTIP_LOCATION.b);
		else
			GameTooltip:AddLine(cItems[index].name, cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b);
		end
	elseif (key == CONVERSE_INDEX_PARTY) and (numparty > 0) then
		GameTooltip:AddDoubleLine(cItems[index].name,
		                          string.format(CONVERSE_TOOLTIP_FORMAT_PARTY, numparty),
		                          cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b,
		                          CONVERSE_COLOUR_TOOLTIP_PLAYERS.r, CONVERSE_COLOUR_TOOLTIP_PLAYERS.g, CONVERSE_COLOUR_TOOLTIP_PLAYERS.b);
	elseif (key == CONVERSE_INDEX_RAID) and (numraid > 0) then
		GameTooltip:AddDoubleLine(cItems[index].name,
		                          string.format(CONVERSE_TOOLTIP_FORMAT_RAID, numraid),
		                          cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b,
		                          CONVERSE_COLOUR_TOOLTIP_PLAYERS.r, CONVERSE_COLOUR_TOOLTIP_PLAYERS.g, CONVERSE_COLOUR_TOOLTIP_PLAYERS.b);
	else
		-- Fallback
		GameTooltip:AddLine(cItems[index].name, cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b);
	end


	-- Show number of lines
	if ConverseHistory[RealmName][key].activity then
		newlines = ConverseHistory[RealmName][key].activity;
	end

	GameTooltip:AddLine(string.format(CONVERSE_TOOLTIP_FORMAT_LINES,
	                    table.getn(ConverseHistory[RealmName][key].text),
	                    ConverseState[RealmName][PlayerName].maxlines, newlines),
			    CONVERSE_COLOUR_TOOLTIP_LINES.r, CONVERSE_COLOUR_TOOLTIP_LINES.g, CONVERSE_COLOUR_TOOLTIP_LINES.b , 1)

	if ConverseHistory[RealmName][key].lastdate then
		GameTooltip:AddLine(string.format(CONVERSE_TOOLTIP_FORMAT_LASTDATE, ConverseHistory[RealmName][key].lastdate),
		                    CONVERSE_COLOUR_TOOLTIP_LASTDATE.r, CONVERSE_COLOUR_TOOLTIP_LASTDATE.g, CONVERSE_COLOUR_TOOLTIP_LASTDATE.b, 1);
	end

	GameTooltip:Show();
end

--=== History ==============================================================--

function ConverseHistory_SetFontSize(size)
	local fontfile, _, fontflags;
	-- History frame
	fontfile, _, fontflags = ConverseFrameHistoryScrollFrame:GetFont();
	ConverseFrameHistoryScrollFrame:SetFont(fontfile, size, fontflags);


	-- Multipaste frame
	fontfile, _, fontflags = ConverseFrameMultipasteFrameScrollFrameText:GetFont();
	ConverseFrameMultipasteFrameScrollFrameText:SetFont(fontfile, size, fontflags);

end

function ConverseHistory_Clear()
	local prefix	= string.sub(cItems[cSelected].key, 1, 2);
	local lname	= string.lower(string.sub(cItems[cSelected].key, 3));
	-- If it's a player or a channel we are not in anymore, remove them from the list
	if (prefix == CONVERSE_PREFIX_MSG) or ((prefix == CONVERSE_PREFIX_CHANNEL) and (not cChannels[lname])) then
		ConverseHistory[RealmName][cItems[cSelected].key] = nil;

		-- Reset to default item
		ConverseItems_Rebuild();

		cSelected = 1;
		if (cNumItems > 0) then
			ConverseState[RealmName][PlayerName].lastselected = cItems[cSelected].key;
			ConverseHistory_Fillup(ConverseState[RealmName][PlayerName].lastselected);
		else
			ConverseFrameHistoryScrollFrame:Clear();
		end

		ConverseItems_Update();
	else
		ConverseHistory[RealmName][cItems[cSelected].key].text = {};
		ConverseHistory[RealmName][cItems[cSelected].key].datestamp = nil;
		ConverseFrameHistoryScrollFrame:Clear();
	end
end


function ConverseHistory_Fillup(key)
	local i;
	ConverseFrameHistoryScrollFrame:Clear();
	if (cNumItems > 0) and (cSelected > 0) and ConverseHistory[RealmName][key] then
		if ConverseHistory[RealmName][key] and ConverseHistory[RealmName][key].text then
			for i = 1, table.getn(ConverseHistory[RealmName][key].text) do
				ConverseFrameHistoryScrollFrame:AddMessage(ConverseHistory[RealmName][key].text[i], cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b ) 
			end
		end
	end
	if (ConverseState[RealmName][PlayerName].maxlines > 0) then
		ConverseFrameLinesText:SetText(ConverseFrameHistoryScrollFrame:GetNumMessages().."/"..ConverseState[RealmName][PlayerName].maxlines);
	else
		ConverseFrameLinesText:SetText(ConverseFrameHistoryScrollFrame:GetNumMessages()..CONVERSE_HEADER_LINES);
	end

end


--=== Ignores ==============================================================--

function ConverseIgnore_OnClick(self, button)
	local checked	= self:GetChecked();
	local id	= self:GetID();
	local key;

	if (id == CONVERSE_IGNORE_ID_WHISPER) then
		key = CONVERSE_IGNORE_WHISPER;
	elseif (id == CONVERSE_IGNORE_ID_SAY) then
		key = CONVERSE_IGNORE_SAY;
	elseif (id == CONVERSE_IGNORE_ID_PARTY) then
		key = CONVERSE_IGNORE_PARTY;
	elseif (id == CONVERSE_IGNORE_ID_RAID) then
		key = CONVERSE_IGNORE_RAID;
	elseif (id == CONVERSE_IGNORE_ID_INS) then
		key = CONVERSE_IGNORE_INS;
	elseif (id == CONVERSE_IGNORE_ID_GUILD) then
		key = CONVERSE_IGNORE_GUILD;
	elseif (id == CONVERSE_IGNORE_ID_OFFICER) then
		key = CONVERSE_IGNORE_OFFICER;
	else
		key = id;
	end

	if (checked) then
		ConverseState[RealmName][PlayerName].ignore[key] = 1;
		PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
	else
		ConverseState[RealmName][PlayerName].ignore[key] = nil;
		PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_OFF);
	end

	-- Refresh

	-- FIXME: self IS CURRENTLY NOT WORKING!

	local cnum,cname = GetChannelName(id);
	if (cnum > 0 and id == cnum) then
		cSelected = 1;
		ConverseState[RealmName][PlayerName].lastselected = CONVERSE_DEF_SELECTED;
	end	

	ConverseItems_Rebuild();
	ConverseItems_Update();

	ConverseHistory_Fillup(ConverseState[RealmName][PlayerName].lastselected);
end


--=== Window Handling ======================================================--

function Converse_Toggle()
	if (ConverseFrame:IsVisible()) then
		ConverseFrame:Hide();
		ConverseState[RealmName][PlayerName].hidden = 1;
		PlaySound(SOUNDKIT.IG_QUEST_LIST_OPEN);
	else
		ConverseFrame:Show();
		ConverseState[RealmName][PlayerName].hidden = nil;
		PlaySound(SOUNDKIT.IG_QUEST_LIST_CLOSE);
		ConverseHistory_Fillup(cItems[cSelected].key);
	end
end

function Converse_OnDragStart()
	if not ConverseState[RealmName][PlayerName].locked then
		ConverseFrame:StartMoving();
	end
end

function Converse_OnDragStop()
	ConverseFrame:StopMovingOrSizing();
end

function ConverseLock_Toggle(self)
	local checked = self:GetChecked();
	if (checked) then
		ConverseState[RealmName][PlayerName].locked = 1;
		PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_OFF);
	else
		ConverseState[RealmName][PlayerName].locked = nil;
		PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
	end
end


--=== Multi-line Paste ======================================================--

function ConverseMultipaste_Toggle()
	PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
	if ConverseFrameMultipasteFrame:IsVisible() then
		ConverseFrameMultipasteFrame:Hide();
		ConverseFrameMultipasteFrameScrollFrameText:SetText("")  
		ConverseFrameMultipasteFrameScrollFrameText:ClearFocus();
	else

		local hfselected = cItems[cSelected].key
		local hftext = ""

		if (cNumItems > 0) and (cSelected > 0) and ConverseHistory[RealmName][hfselected] then

			local hfi = 0
			local hfcolor = cColourString(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b)
		
			if ConverseHistory[RealmName][hfselected] and ConverseHistory[RealmName][hfselected].text then
				for hfi = 1 , table.getn(ConverseHistory[RealmName][hfselected].text) do
					hftext = hftext .. hfcolor .. ConverseHistory[RealmName][hfselected].text[hfi] .."\n"
				end
			end
		end
		ConverseFrameMultipasteFrameScrollFrameText:SetText(hftext);
		ConverseFrameMultipasteFrame:Show();
--		ConverseFrameMultipasteFrameScrollFrameText:SetFocus();
	end
end

function ConverseMultipaste_OnTextChanged()
	if cFinishedLoading then
		ConverseState[RealmName][PlayerName].multipaste = ConverseFrameMultipasteFrameScrollFrameText:GetText();
	end
end
